"""Testing the tensorflow adaptor"""
from unittest import TestCase
import tensorflow as tf
import shutil
import os

from dlhub_sdk.models.servables.tensorflow import TensorFlowModel
from dlhub_sdk.utils.schemas import validate_against_dlhub_schema

tf_export_path = os.path.join(os.path.dirname(__file__), 'tf-model')


class TestTensorflow(TestCase):

    maxDiff = 4096

    def setUp(self):
        # Clear existing model
        if os.path.isdir(tf_export_path):
            shutil.rmtree(tf_export_path)

    def make_model(self):

        tf.reset_default_graph()

        with tf.Session() as sess:

            # Make two simple graphs, both of which will be served by TF
            x = tf.placeholder('float', shape=(None, 3), name='Input')
            z = tf.placeholder('float', shape=(), name='ScalarMultiple')
            y = x + 1
            len_fun = tf.reduce_sum(y - x)  # Returns the number of elements in the array
            scale_mult = tf.multiply(z, x, name='scale_mult')

            # Create the tool for saving the model to disk
            builder = tf.saved_model.builder.SavedModelBuilder(tf_export_path)

            #  Make descriptions for the inputs and outputs
            x_desc = tf.saved_model.utils.build_tensor_info(x)
            y_desc = tf.saved_model.utils.build_tensor_info(y)
            z_desc = tf.saved_model.utils.build_tensor_info(z)
            len_fun_desc = tf.saved_model.utils.build_tensor_info(len_fun)
            scale_mult_desc = tf.saved_model.utils.build_tensor_info(scale_mult)

            #  Make a signature for the functions to be served
            func_sig = tf.saved_model.signature_def_utils.build_signature_def(
                inputs={'x': x_desc},
                outputs={'y': y_desc},
                method_name='run'
            )
            len_sig = tf.saved_model.signature_def_utils.build_signature_def(
                inputs={'x': x_desc},
                outputs={'len': len_fun_desc},
                method_name='length'
            )
            mult_sig = tf.saved_model.signature_def_utils.build_signature_def(
                inputs={'x': x_desc, 'z': z_desc},
                outputs={'scale_mult': scale_mult_desc},
                method_name='scalar_multiply'
            )

            #  Add the functions and the  state of the graph to the builder
            builder.add_meta_graph_and_variables(
                sess, [tf.saved_model.tag_constants.SERVING],
                signature_def_map={
                    tf.saved_model.signature_constants.DEFAULT_SERVING_SIGNATURE_DEF_KEY: func_sig,
                    'length': len_sig,
                    'scalar_multiply': mult_sig
                })

            #  Save the function
            builder.save()

    def test_tf(self):
        # Make a model and save it to disk
        self.make_model()

        # Create the description
        model = TensorFlowModel.create_model(tf_export_path).set_title('TF Test')\
            .set_name('tf-test')

        # Generate the metadata for the test
        metadata = model.to_dict(simplify_paths=True)

        # Check whether the 'x' is listed first for the multiple-input model or second
        self.assertEqual({'other': ['saved_model.pb']}, metadata['dlhub']['files'])
        self.assertEqual(metadata['servable'],
                         {'methods':
                             {'run': {
                                 'input': {'type': 'ndarray', 'description': 'x',
                                           'shape': [None, 3], 'item_type': {'type': 'float'}},
                                 'output': {'type': 'ndarray', 'description': 'y',
                                            'shape': [None, 3], 'item_type': {'type': 'float'}},
                                 'parameters': {},
                                 'method_details': {'input_nodes': ['Input:0'],
                                                    'output_nodes': ['add:0']}
                             }, 'length': {
                                 'input': {'type': 'ndarray', 'description': 'x',
                                           'shape': [None, 3], 'item_type': {'type': 'float'}},
                                 'output': {'type': 'float', 'description': 'len'},
                                 'parameters': {},
                                 'method_details': {'input_nodes': ['Input:0'],
                                                    'output_nodes': ['Sum:0']}
                             }, 'scalar_multiply': {
                                 'input': {'type': 'list', 'description': 'Arguments',
                                           'item_type': [
                                               {'type': 'ndarray', 'description': 'x',
                                                'shape': [None, 3], 'item_type': {'type': 'float'}},
                                               {'type': 'float', 'description': 'z'}
                                           ]},
                                 'output': {'type': 'ndarray', 'description': 'scale_mult',
                                            'shape': [None, 3], 'item_type': {'type': 'float'}},
                                 'parameters': {},
                                 'method_details': {'input_nodes': ['Input:0', 'ScalarMultiple:0'],
                                                    'output_nodes': ['scale_mult:0']}
                             }},
                             'shim': 'tensorflow.TensorFlowServable',
                             'type': 'TensorFlow Model',
                             'dependencies': {'python': {'tensorflow': tf.__version__}}})

        validate_against_dlhub_schema(metadata, 'servable')



from dlhub_sdk.models.servables.python import PythonStaticMethodModel
from dlhub_sdk.client import DLHubClient
from unittest import TestCase, skip
import pandas as pd


class TestClient(TestCase):

    def test_dlhub_init(self):
        dl = DLHubClient()
        self.assertEqual(dl.service, "https://api.dlhub.org/api/v1")
        self.assertIsInstance(dl, DLHubClient)

    def test_get_servables(self):
        dl = DLHubClient()
        r = dl.get_servables()
        self.assertIsInstance(r, pd.DataFrame)
        self.assertGreater(r.shape[-1], 0)
        self.assertNotEqual(r.shape[0], 0)

    def test_get_id_by_name(self):
        dl = DLHubClient()
        name = "noop"
        r = dl.get_id_by_name(name)
        r2 = dl.get_servables()
        true_val = r2["uuid"][r2["name"].tolist().index(name)]
        self.assertEqual(r, true_val)

        # Invalid name
        with self.assertRaises(IndexError):
            dl.get_id_by_name("foo")

    def test_run(self):
        dl = DLHubClient(timeout=10)
        name = "noop"
        data = {"data": ["V", "Co", "Zr"]}
        serv = dl.get_id_by_name(name)
        res = dl.run(serv, data)

        # Check the results
        self.assertIsInstance(res, pd.DataFrame)

    @skip  # Do not yet have a "test" route for submitted objects
    def test_submit(self):
        dl = DLHubClient()

        # TODO: Set the same UUID as the original test

        # Make an example function
        model = PythonStaticMethodModel.create_model('numpy.linalg', 'norm')
        model.set_name('1d_norm')
        model.set_title('Norm of a 1D Array')
        model.set_inputs('ndarray', 'Array to be normed', shape=[None])
        model.set_outputs('number', 'Norm of the array')

        # Submit the model
        dl.publish_servable(model)



import unittest

from dlhub_sdk.models.servables.python import PythonStaticMethodModel
from dlhub_sdk.utils.schemas import validate_against_dlhub_schema
from dlhub_sdk.models.pipeline import PipelineModel
from dlhub_sdk import __version__


class TestPipeline(unittest.TestCase):

    def test_pipeline(self):
        """Make a pipeline composed of two numpy steps"""

        # Generate the two steps
        step1 = PythonStaticMethodModel.create_model('numpy', 'max', function_kwargs={'axis': 1})
        step2 = PythonStaticMethodModel.create_model('numpy', 'mean')

        # Assign them dlhub IDs
        step1.assign_dlhub_id()
        step2.assign_dlhub_id()

        # Make the pipeline
        pipeline = PipelineModel().set_title('Average of Column Maximums').set_name('numpy_test')
        pipeline.add_step(step1.dlhub_id, 'Maximum of each column', {'axis': 0})
        pipeline.add_step(step2, 'Average of the maximums')

        # Generate the pipeline metadata
        metadata = pipeline.to_dict()
        self.assertEqual(metadata,
            {'datacite': {'creators': [], 'titles': [{'title': 'Average of Column Maximums'}],
                          'publisher': 'DLHub', 'publicationYear': '2018',
                          'identifier': {'identifier': '10.YET/UNASSIGNED',
                                         'identifierType': 'DOI'},
                          'resourceType': {'resourceTypeGeneral': 'InteractiveResource'},
                          "descriptions": [],
                          "fundingReferences": [],
                          "relatedIdentifiers": [],
                          "alternateIdentifiers": [],
                          "rightsList": []},
             'dlhub': {'version': __version__, 'domains': [], 'visible_to': ['public'],
                       'id': None, 'name': 'numpy_test',
                       'files': {}},
             'pipeline': {'steps': [{'dlhub_id': step1.dlhub_id,
                           'description': 'Maximum of each column', 'parameters': {'axis': 0}},
                          {'dlhub_id': step2.dlhub_id,
                           'description': 'Average of the maximums'}]}})
        validate_against_dlhub_schema(metadata, 'pipeline')



"""Utilities for validating against DLHub schemas"""
from jsonschema import Draft4Validator, RefResolver
import requests

_schema_repo = "https://raw.githubusercontent.com/DLHub-Argonne/dlhub_schemas/master/schemas/"


def validate_against_dlhub_schema(document, schema_name):
    """Validate a metadata document against one of the DLHub schemas

    Note: Requires an internet connection

    Args:
        document (dict): Document instance to be validated
        schema_name (string): Name of schema (e.g., "dataset" for validating datasets).
            For full list, see: https://github.com/DLHub-Argonne/dlhub_schemas
    Raises:
        (jsonschema.SchemaError) If the schema fails to validate
    """

    # Make the schema validator
    schema = requests.get("{}/{}.json".format(_schema_repo, schema_name)).json()
    validator = Draft4Validator(schema, resolver=RefResolver(_schema_repo, schema))

    # Test the document
    validator.validate(document)



from datetime import datetime
import numpy as np
import unittest
import os

from dlhub_sdk.models.servables.sklearn import ScikitLearnModel
from dlhub_sdk import __version__
from dlhub_sdk.utils.schemas import validate_against_dlhub_schema


_year = str(datetime.now().year)


class TestSklearn(unittest.TestCase):
    maxDiff = 4096

    def test_load_model(self):
        model_path = os.path.abspath(os.path.join(os.path.dirname(__file__), 'model.pkl'))

        # Load the model
        model_info = ScikitLearnModel.create_model(model_path, n_input_columns=4, classes=3)
        model_info.set_title('Sklearn example').set_name('sklearn')
        expected = {'datacite': {'creators': [], 'publisher': 'DLHub',
                                 'titles': [{'title': 'Sklearn example'}],
                                 'resourceType': {'resourceTypeGeneral': "InteractiveResource"},
                                 'identifier': {'identifier': '10.YET/UNASSIGNED',
                                                'identifierType': 'DOI'},
                                 'publicationYear': _year,
                                 "descriptions": [],
                                 "fundingReferences": [],
                                 "relatedIdentifiers": [],
                                 "alternateIdentifiers": [],
                                 "rightsList": []},
                    "dlhub": {"version": __version__,
                              "visible_to": ["public"], 'name': 'sklearn',
                              "domains": [], "id": None,
                              'files': {'model': model_path}},
                    'servable': {
                        'language': 'python',
                        'type': 'Scikit-learn estimator',
                        'dependencies': {
                            'python': {'scikit-learn': '0.19.1'}
                        },
                        'model_type': 'SVC',
                        'shim': 'sklearn.ScikitLearnServable',
                        'methods': {
                            'run': {"input": {
                                    "type": "ndarray",
                                    "shape": [None, 4],
                                    "description": "List of records to evaluate with model. Each record is a list of 4 variables.",
                                    "item_type": {"type": "float"}
                                }, "output": {
                                    "type": "ndarray",
                                    "shape": [None, 3],
                                    "description": "Probabilities for membership in each of 3 classes",
                                    "item_type": {"type": "float"}
                                },
                                "parameters": {},
                                "method_details": {
                                    "method_name": "_predict_proba"
                                }
                            }
                        },
                        'model_summary': "SVC(C=1, cache_size=200, class_weight=None, coef0=0.0,\n"
                                         "  decision_function_shape='ovr', degree=3, gamma='auto', kernel='linear',\n"
                                         "  max_iter=-1, probability=True, random_state=None, shrinking=True,\n"
                                         "  tol=0.001, verbose=False)",
                        'options': {
                            'is_classifier': True,
                            'serialization_method': 'pickle',
                            'classes': ['Class 1', 'Class 2', 'Class 3']
                        }
                    }}
        self.assertEquals(model_info.to_dict(), expected)
        validate_against_dlhub_schema(model_info.to_dict(), 'servable')

    def test_regression(self):
        """Test a regression model saved with joblib"""
        model_path = os.path.abspath(os.path.join(os.path.dirname(__file__), 'model-lr.pkl'))

        # Load the model
        model_info = ScikitLearnModel.create_model(model_path, n_input_columns=2,
                                                   serialization_method='joblib',
                                                   classes=np.array(['number']))

        # Check that the metadata is as expected
        self.assertEqual(model_info["servable"]["methods"]["run"]["method_details"]["method_name"],
                         "predict")
        self.assertEqual([model_path], model_info.list_files())
        self.assertEqual(['number'], model_info["servable"]["options"]["classes"])
        self.assertEqual([None], model_info["servable"]["methods"]["run"]['output']['shape'])



from dlhub_sdk.utils.schemas import validate_against_dlhub_schema
from dlhub_sdk.models.servables.keras import KerasModel
from dlhub_sdk import __version__
from keras import __version__ as keras_version
from keras.models import Sequential, Model
from keras.layers import Dense, Input
from h5py import __version__ as h5py_version
from unittest import TestCase
from tempfile import mkdtemp
import shutil
import os


class TestKeras(TestCase):

    maxDiff = 4096

    def test_keras_single_input(self):
        # Make a Keras model
        model = Sequential()
        model.add(Dense(16, input_shape=(1,), activation='relu', name='hidden'))
        model.add(Dense(1, name='output'))
        model.compile(optimizer='rmsprop', loss='mse')

        # Save it to disk
        tempdir = mkdtemp()
        try:
            model_path = os.path.join(tempdir, 'model.hd5')
            model.save(model_path)

            # Create a model
            metadata = KerasModel.create_model(model_path, ["y"])
            metadata.set_title('Keras Test')
            metadata.set_name('mlp')

            output = metadata.to_dict()
            self.assertEqual(output, {
                "datacite": {"creators": [], "titles": [{"title": "Keras Test"}],
                             "publisher": "DLHub", "publicationYear": "2018",
                             "identifier": {"identifier": "10.YET/UNASSIGNED",
                                            "identifierType": "DOI"},
                             "resourceType": {"resourceTypeGeneral": "InteractiveResource"},
                             "descriptions": [],
                             "fundingReferences": [],
                             "relatedIdentifiers": [],
                             "alternateIdentifiers": [],
                             "rightsList": []},
                "dlhub": {"version": __version__, "domains": [],
                          "visible_to": ["public"], "id": None,
                          "name": "mlp", "files": {"model": model_path}},
                "servable": {"methods": {"run": {
                    "input": {"type": "ndarray", "description": "Tensor", "shape": [None, 1]},
                    "output": {"type": "ndarray", "description": "Tensor",
                               "shape": [None, 1]}, "parameters": {},
                    "method_details": {
                        "method_name": "predict",
                        "classes": ["y"]
                    }}},
                    "type": "Keras Model",
                    "shim": "keras.KerasServable",
                    "language": "python",
                    "model_type": "Deep NN",
                    "model_summary": """_________________________________________________________________
Layer (type)                 Output Shape              Param #   
=================================================================
hidden (Dense)               (None, 16)                32        
_________________________________________________________________
output (Dense)               (None, 1)                 17        
=================================================================
Total params: 49
Trainable params: 49
Non-trainable params: 0
_________________________________________________________________
""",
                    "dependencies": {"python": {
                        'keras': keras_version,
                        'h5py': h5py_version
                    }}}})

            # Validate against schema
            validate_against_dlhub_schema(output, 'servable')
        finally:
            shutil.rmtree(tempdir)

    def test_keras_multioutput(self):
        # Make a Keras model
        input_layer = Input(shape=(4,))
        dense = Dense(16, activation='relu')(input_layer)
        output_1 = Dense(1, activation='relu')(dense)
        output_2 = Dense(2, activation='softmax')(dense)
        model = Model([input_layer], [output_1, output_2])
        model.compile(optimizer='rmsprop', loss='mse')

        # Save it to disk
        tempdir = mkdtemp()
        try:
            model_path = os.path.join(tempdir, 'model.hd5')
            model.save(model_path)

            # Create a model
            metadata = KerasModel.create_model(model_path, [['y'], ['yes', 'no']])
            metadata.set_title('Keras Test')
            metadata.set_name('mlp')

            self.assertEqual(metadata['servable']['methods']['run']['output'],
                             {'type': 'list',
                              'description': 'List of tensors',
                              'item_type': [
                                  {'type': 'ndarray', 'description': 'Tensor', 'shape': [None, 1]},
                                  {'type': 'ndarray', 'description': 'Tensor', 'shape': [None, 2]}
                              ]})

            output = metadata.to_dict()

            # Validate against schema
            validate_against_dlhub_schema(output, 'servable')
        finally:
            shutil.rmtree(tempdir)



from dlhub_sdk.models.datasets import Dataset, TabularDataset
from dlhub_sdk import __version__

from datetime import datetime
from tempfile import mkstemp
from zipfile import ZipFile
from glob import glob
import unittest
import uuid
import os

from dlhub_sdk.utils.schemas import validate_against_dlhub_schema


_year = str(datetime.now().year)


class TestModels(unittest.TestCase):

    maxDiff = None

    def test_add_directory(self):
        # Get how many files in directory below this
        my_dir = os.path.dirname(__file__)
        my_count = sum([os.path.isfile(x) for x in glob(my_dir + '/*')])

        # Test it out
        self.assertEquals(my_count, len(Dataset().add_directory(my_dir).list_files()))
        self.assertLessEqual(my_count,
                             len(Dataset().add_directory(my_dir, recursive=True).list_files()))

    def test_dataset(self):
        my_uuid = str(uuid.uuid1())
        m = Dataset().set_authors(["Ward, Logan"], ["University of Chicago"])\
            .set_title("Example dataset").add_alternate_identifier("10.11", "DOI")\
            .add_related_identifier("10.11", "DOI", 'IsDescribedBy')\
            .add_funding_reference("ANL LDRD", '1', 'ISNI', '201801', 'DLHub', 'http://funding.uri')\
            .set_version(1)\
            .add_rights("https://www.gnu.org/licenses/gpl-3.0.en.html", "GPL v3.0")\
            .set_abstract("Abstract").set_methods("Methods")\
            .set_visibility(['public']).set_domains(["materials science"])\
            .set_dlhub_id(my_uuid).set_name("example_data")
        self.assertEqual(m.to_dict(),
                         {"datacite":
                              {"creators": [{"givenName": "Logan", "familyName": "Ward",
                                             "affiliations": "University of Chicago"}],
                               "titles": [{'title': "Example dataset"}],
                               "publisher": 'DLHub',
                               "publicationYear": _year,
                               "version": '1',
                               "resourceType": {"resourceTypeGeneral": "Dataset"},
                               "descriptions": [{
                                   "description": "Abstract", "descriptionType": "Abstract"
                               }, {
                                   "description": "Methods", "descriptionType": "Methods"
                               }],
                               "fundingReferences": [{
                                   "awardNumber": {"awardNumber": "201801",
                                                   "awardURI": "http://funding.uri"},
                                   "awardTitle": "DLHub",
                                   "funderIdentifier": {'funderIdentifier': '1',
                                                        'funderIdentifierType': 'ISNI'},
                                   "funderName": "ANL LDRD"
                               }],
                               "relatedIdentifiers": [{
                                   "relatedIdentifier": "10.11",
                                   "relatedIdentifierType": "DOI",
                                   "relationType": "IsDescribedBy"
                               }],
                               "alternateIdentifiers": [{
                                   "alternateIdentifier": "10.11",
                                   "alternateIdentifierType": "DOI"
                               }],
                               "rightsList": [{
                                   "rightsURI": "https://www.gnu.org/licenses/gpl-3.0.en.html",
                                   "rights": "GPL v3.0"
                               }],
                               'identifier': {'identifier': '10.YET/UNASSIGNED',
                                              'identifierType': 'DOI'},
                               },
                          "dlhub": {
                              "version": __version__,
                              "visible_to": ["public"],
                              "domains": ["materials science"],
                              "id": my_uuid,
                              "name": "example_data",
                              "files": {}
                          }, "dataset": {}})
        validate_against_dlhub_schema(m.to_dict(), "dataset")

    def test_tabular_dataset(self):
        data_path = os.path.abspath(os.path.join(os.path.dirname(__file__), 'test.csv'))
        m = TabularDataset.create_model(data_path)

        # Add some nonsense
        m.set_title('Example dataset')
        m.set_name('example_dataset')
        with self.assertRaises(ValueError):
            m.set_name('has whitespace')
        m.mark_inputs(['x'])
        m.mark_labels(['y'])
        m.annotate_column('x', description='Input variable', units='cm')
        m.annotate_column('y', data_type='scalar')
        self.assertEqual(m.to_dict(), {"datacite": {"titles": [{'title': "Example dataset"}],
                                                    "creators": [], "publisher": "DLHub",
                                                    "resourceType": {
                                                        "resourceTypeGeneral": "Dataset"
                                                    }, "publicationYear": _year,
                                                    'identifier': {
                                                        'identifier': '10.YET/UNASSIGNED',
                                                        'identifierType': 'DOI'},
                                                    "descriptions": [],
                                                    "fundingReferences": [],
                                                    "relatedIdentifiers": [],
                                                    "alternateIdentifiers": [],
                                                    "rightsList": [],
                                                    },
                                       "dlhub": {"version": __version__,
                                                 "visible_to": ["public"],
                                                 "domains": [],
                                                 "id": None,
                                                 "name": "example_dataset",
                                                 "files": {'data': data_path}},
                                       "dataset": {"columns": [
                                                       {"name": "x",
                                                        "description": "Input variable",
                                                        "type": "integer", "units": "cm"},
                                                       {"name": "y", "type": "scalar"}],
                                                   "inputs": ["x"], "labels": ["y"],
                                                   "format": "csv", "read_options": {}}})
        validate_against_dlhub_schema(m.to_dict(), "dataset")

        # Test the simplification of files
        metadata = m.to_dict(simplify_paths=True)
        self.assertEqual({'data': 'test.csv'}, metadata['dlhub']['files'])

    def test_zip(self):
        """Test generating a zip file with the requested files"""

        data_path = os.path.abspath(os.path.join(os.path.dirname(__file__), 'test.csv'))
        m = TabularDataset.create_model(data_path)

        # Create a temp file for the ZIP
        fp, temp_path = mkstemp(".zip")
        os.close(fp)
        os.unlink(temp_path)

        try:
            # Test making the ZIP file with just the dataset
            cp = m.get_zip_file(temp_path)
            self.assertEqual(os.path.abspath(os.path.dirname(data_path)), cp)

            # Make sure it has only one file ('test.csv')
            with ZipFile(temp_path) as zf:
                z_files = set(f.filename for f in zf.filelist)
                self.assertEqual({'test.csv'}, z_files)

            # Delete the ZIP file (we are in exclusive create mode)
            os.unlink(temp_path)

            # Add the `dataset.py` file in the directory below this one
            m.add_files(os.path.join(os.path.dirname(data_path), '..', 'datasets.py'))

            m.get_zip_file(temp_path)
            with ZipFile(temp_path) as zf:
                z_files = set(f.filename for f in zf.filelist)
                self.assertEqual({'datasets.py', 'tests/test.csv'}, z_files)

            self.assertEqual(os.path.abspath(os.path.dirname(data_path)), cp)

            # Test an empty ZIP file
            os.unlink(temp_path)

            cp = Dataset().get_zip_file(temp_path)
            self.assertEqual(cp, '.')
            with ZipFile(temp_path) as zf:
                self.assertEqual(0, len(zf.filelist))

        finally:
            os.unlink(temp_path)

    def test_serialize(self):
        # Make metadata where I overwrite a auto-generated field
        metadata = Dataset().set_title('Test').set_name('test')
        metadata['dlhub']['version'] = 'test'

        # Pickle/unpickle the object
        metadata_copy = Dataset.from_dict(metadata.to_dict())

        # Make sure the fields are the same
        self.assertEqual(metadata._output, metadata_copy._output)

    def test_unlabeled(self):
        data_path = os.path.abspath(os.path.join(os.path.dirname(__file__), 'test.csv'))
        m = TabularDataset.create_model(data_path)

        # Check that all columns are unlabeled
        self.assertEqual(["x", "y"], m.get_unannotated_columns())

        # Label one, make sure it changes
        m.annotate_column("x", "description")
        self.assertEqual(["y"], m.get_unannotated_columns())

if __name__ == "__main__":
    unittest.main()



"""Tests for models for generic Python functions"""

from dlhub_sdk.models.servables.python import PythonClassMethodModel, \
    PythonStaticMethodModel
from dlhub_sdk import __version__
from sklearn import __version__ as skl_version
from numpy import __version__ as numpy_version
from datetime import datetime
import unittest
import math
import os

from dlhub_sdk.utils.schemas import validate_against_dlhub_schema


_year = str(datetime.now().year)


class TestPythonModels(unittest.TestCase):

    maxDiff = 4096

    def test_pickle(self):
        pickle_path = os.path.abspath(os.path.join(os.path.dirname(__file__), 'model.pkl'))

        # Make the model
        model = PythonClassMethodModel.create_model(pickle_path, 'predict_proba', {'fake': 'kwarg'})
        model.set_title('Python example').set_name("class_method")

        # Make sure it throws value errors if inputs are not set
        with self.assertRaises(ValueError):
            model.to_dict()

        # Define the input and output types
        model.set_inputs('ndarray', 'Features for each entry', shape=[None, 4])
        model.set_outputs('ndarray', 'Predicted probabilities of being each iris species',
                          shape=[None, 3])

        # Add some requirements
        model.add_requirement('scikit-learn', 'detect')
        model.add_requirement('numpy', 'detect')
        model.add_requirement('theano', 'latest')  # Deprecated project, version should stay same

        # Check the model output
        output = model.to_dict()
        self.assertEqual(output,
                         {'datacite':
                              {'creators': [], 'titles': [{'title': 'Python example'}],
                               'publisher': 'DLHub',
                               'resourceType': {'resourceTypeGeneral': 'InteractiveResource'},
                               'identifier': {'identifier': '10.YET/UNASSIGNED',
                                              'identifierType': 'DOI'},
                               'publicationYear': _year,
                               "descriptions": [],
                               "fundingReferences": [],
                               "relatedIdentifiers": [],
                               "alternateIdentifiers": [],
                               "rightsList": []
                               },
                          'dlhub': {'version': __version__, 'domains': [],
                                    'visible_to': ['public'],
                                    "id": None, "name": "class_method",
                                    'files': {'pickle': pickle_path}},
                          'servable': {'language': 'python', 'type': 'Python class method',
                                       'shim': 'python.PythonClassMethodServable',
                                       'methods': {'run': {'input': {'type': 'ndarray',
                                                                     'description': 'Features for each entry',
                                                                     'shape': [None, 4]},
                                                           'output': {'type': 'ndarray',
                                                                      'description': 'Predicted probabilities of being each iris species',
                                                                      'shape': [None, 3]},
                                                           'parameters': {'fake': 'kwarg'},
                                                           'method_details': {
                                                               'class_name': 'sklearn.svm.classes.SVC',
                                                               'method_name': 'predict_proba'},
                                                           }},
                                       'dependencies': {
                                           'python': {
                                               'scikit-learn': skl_version,
                                               'numpy': numpy_version,
                                               'theano': '1.0.3'
                                           }}
                          }})
        self.assertEqual([pickle_path], model.list_files())
        validate_against_dlhub_schema(output, 'servable')

    def test_function(self):
        f = math.sqrt

        # Make the model
        model = PythonStaticMethodModel.from_function_pointer(f, autobatch=True)
        model.set_name("static_method").set_title('Python example')

        # Describe the inputs/outputs
        model.set_inputs('list', 'List of numbers', item_type='float')
        model.set_outputs('float', 'Square root of the number')

        # Generate the output
        output = model.to_dict()
        self.assertEqual(output,
                         {'datacite':
                              {'creators': [], 'titles': [{'title': 'Python example'}],
                               'publisher': 'DLHub',
                               'resourceType': {'resourceTypeGeneral': 'InteractiveResource'},
                               'identifier': {'identifier': '10.YET/UNASSIGNED',
                                              'identifierType': 'DOI'},
                               'publicationYear': _year,
                               "descriptions": [],
                               "fundingReferences": [],
                               "relatedIdentifiers": [],
                               "alternateIdentifiers": [],
                               "rightsList": []
                           },
                          'dlhub': {'version': __version__, 'domains': [], 'visible_to': ['public'],
                                    "id": None, "name": "static_method",
                                    'files': {}},
                          'servable': {'language': 'python', 'type': 'Python static method',
                                       'shim': 'python.PythonStaticMethodServable',
                                       'methods': {'run': {'input': {'type': 'list',
                                                                     'description': 'List of numbers',
                                                                     'item_type': {
                                                                         'type': 'float'}},
                                                           'output': {'type': 'float',
                                                                      'description': 'Square root of the number'},
                                                           'parameters': {},
                                                           'method_details': {'module': 'math',
                                                                              'method_name': 'sqrt',
                                                                              'autobatch': True}}
                                                   },
                                       'dependencies': {'python': {}}}
                          })
        validate_against_dlhub_schema(output, 'servable')
